#!/usr/bin/perl -w
# Copyright (c) 2010 - 2012 University of Cambridge
# This code is provided under licence with you and your institution/company.
# All rights reserved.
#
# Refer to the Licence.txt file in your distribution kit.
#
# FETA (Wizard v2.52 17/01/2013)
# Implemented by David Parry-Smith
# CNC Programmer
# University of Cambridge
#
# This file must NEVER be edited - it is NOT the source file!
# This is a literate program, you must edit the ffqrun2.w file and
# run the literate commands:
# nuweb ffqrun2
# pdflatex ffqrun2
# (You may need to run that last command at least twice)
# to generate this Perl file and the PDF documentation file.
#
# ffqrun2.w is the source file for this Perl script.
#
# Wondering why there are no comments in this file?
# You should be reading ffqrun2.pdf - that file contains
# the complete commented source. You should definitely not
# be editing this file.



use strict;
use constant TRUE => 1;
use constant FALSE => 0;


BEGIN {
    $ENV{PERL_DL_NONLAZY} = 1;
}
use Tkx;


BEGIN {
    $ENV{PERL_DL_NONLAZY} = 0;
}

use List::Util qw( max );


use Time::localtime;
# my $tm = localtime;
# my $date = sprintf("%04d-%02d-%02dT%02d:%02d:%02d",$tm->year+1900, ($tm->mon)+1, $tm->mday,$tm->hour, $tm->min, $tm->sec);


# Tkx::package_require('tooltip');
# Tkx::namespace_import("::tooltip::tooltip");


my $Main_window;
my $Container;
my $Centre_frame;

my $BACK_button;
my $CANCEL_button;
my $Action_Buttons_row;

my $NEXT_button;
my @Wizard_steps_visited;

my @Widget_refs_to_go;

my @Run_mode_options = qw( spreadsheet nutrients meals foods  );

my %Run_mode_labels = (
    'spreadsheet' => 'in wide format',
    'nutrients' => 'in long format',
    'meals' => 'by ffq line',
    'foods' => 'by food item and ffq line',
);
my %Run_mode_tips = (
    'spreadsheet'
        =>  <<SPREADSHEET,
Average daily nutrient values
for an individual, in wide format,
suitable for import into a spreadsheet
or stats package.
SPREADSHEET
    'nutrients'
        => <<NUTRIENTS,
Average daily nutrient values
for an individual, in long format.
Mostly suitable for programmers
NUTRIENTS
    'meals'
        => <<MEALS,
Average daily nutrient values
for each ffq line, for an individual.
Mostly suitable for programmers.
Expect a very large output file.
MEALS
    'foods'
        => <<FOODS
Average daily nutrient values
for each food code, for each ffq
line,for an individual.
Mostly suitable for programmers.
Expect a very large output file.
FOODS
);

my $Mode_selection = 'spreadsheet';


my $Log_frame;
my $Logging_window;

my $Utility_frame;
my $LICENCE_VIEW_button;
my $EPIC_VIEW_LOG_button;
my $SAVE_button;
my $SETUP_button;

my @Messages_buffer;

my $EPIC_client_program;

my $Epic_ffq_version;

my $Client_config_file_specification;

my $Default_folder;
my $Setup_file_specification;

my $Default_input_directory;
my $Default_output_directory;
my $Default_log_directory;

my $actions = {
    'log_folder' => sub {
        my $verb = shift;
        if ( @_ ) {
            $Default_log_directory = shift;
        }
        return $Default_log_directory;
    },

    'output_folder' => sub {
        my $verb = shift;
        if ( @_ ) {
            $Default_output_directory = shift;
        }
        return $Default_output_directory;
    },

    'input_folder' => sub {
        my $verb = shift;
        if ( @_ ) {
            $Default_input_directory = shift;
        }
        return $Default_input_directory;
    },

    '_DEFAULT_' => sub {
        my $verb = shift;
        confirm(
    "DEFAULT: Unrecognised action $verb on line $. of config file"
        );
        return '';
    },
};


my @Wizard_steps = (
    'step 5',
    'step 4',
    'step 3',
    'step 2',
    'step 1',
);

my %Wizard_step = (
    'step 1'          =>  \&ws_introduction,
    'step 2'          =>  \&ws_select_mode,
    'step 3'          =>  \&ws_get_input_file,
    'step 4'          =>  \&ws_set_output_file,
    'step 5'          =>  \&ws_run_epic_ffq,
    'cancel'          =>  \&ws_cancel_ffq,
);

my %Back_tip = (
    'step 1'          =>  'No prior step',
    'step 2'          =>  'Introduction',
    'step 3'          =>  'Select mode',
    'step 4'          =>  'Input file',
    'step 5'          =>  'Output file',
    'cancel'          =>  'Not available',
);

my %Next_tip = (
    'step 1'          =>  'Select mode',
    'step 2'          =>  'Input file',
    'step 3'          =>  'Output file',
    'step 4'          =>  'Confirm',
    'step 5'          =>  'Finish',
    'cancel'          =>  'Not available',
);

my %Wizard_step_title = (
    'step 1' => 'Welcome to FETA ',
    'step 2' => 'Define FETA client run mode ',
    'step 3' => 'Select input file ',
    'step 4' => 'Specify output file',
    'step 5' => 'Running FETA client... ',
    'cancel' => 'Cancelling FETA ',
);

my %Wizard_step_label = (
'step 1' => <<'STEP1',
Welcome to FETA, the Food frequency questionnaire EPIC Tool for Analysis.

This wizard tool will help you to check the validity of your input file
so that you can successfully process EPIC food frequency questionnaires.

You can go back and forth or cancel at any time. When you are ready, click
the next button to get started.

STEP1
'step 2' => <<'STEP2',
Output format options are presented from highest level
of data aggregation to lowest.

Hover over item text with pointer for more detailed information.

Calculate nutrients for individuals with output:
STEP2
'step 3' => <<'STEP3',
Choose an input file in CSV format.

Please ensure you have read and implemented the guildelines
for creating this file before submitting a job.
STEP3
'step 4' => <<'STEP4',
Choose the location and the name of the output file
to be created in that location.

If you wish to set the location of the log file,
click the Setup button now to set it.

NB. The log file name is generated using the filename you define
here, the date and a serial number with an extension of ".log".

You cannot choose the name of the log file yourself - the progam does this for you.
STEP4
'step 5' => <<'STEP5',
The Wizard is now running your job. Please read any messages
that appear in the window below - they may be important.

Error messages will have important consequences for the validity of your results.
STEP5
'step 3a' => <<'STEP3A',
There are errors in your input file that require your attention
It is best to correct the IDs that have been identified in the messages window below
before submitting your job.
STEP3A
);

my %Back_indicator = (
    'step 1' => 0,
    'step 2' => 1,
    'step 3' => 1,
    'step 4' => 1,
    'step 5' => 0,
    'cancel' => 0,
);

my %Forward_indicator = (
    'step 1' => 1,
    'step 2' => 1,
    'step 3' => 1,
    'step 4' => 1,
    'step 5' => 0,
    'cancel' => 0,
);

my $Input_folder;
my $Input_file_specification;

my $Input_widget;
my $BROWSE_button;

my %participant_q_errors;

my %mult_freq;

my %Count_reds;
my $Red_limit = 1;

my $Step_title;

my $Output_folder;
my $Output_file_specification;

my $Output_widget;

my $Current_step_name;

my @Log_buffer;

my $Setup_feedback_frequency = $ENV{'EPIC_FEEDBACK_FREQUENCY'} || 22;

my $nutrients_file_specification;

my $EPICFFQ_LOGFILE;

my $Log_folder;
my $Log_file_specification;

my $Ffr_file_specification;

my @Licence_buffer;


my $Wizard_frame_title = "FETA (Wizard v2.51 20/8/2012)";
my $Wizard_current_step = 0;
my $Wizard_total_steps = 5;



D_start_interface();

sub D_start_interface {

   $Main_window = Tkx::widget->new(".");






   $Main_window->g_wm_title($Wizard_frame_title);



   $Container = $Main_window->new_ttk__frame(
       -padding=>"10"
   );
   $Container->configure(
       -borderwidth => 3
   );
   $Container->g_grid(
       -column => 0,
       -row => 0,
       -sticky => 'nsew',
   );
   $Centre_frame = $Container->new_ttk__labelframe(
       -text => 'Starting up...',
       -padding => "3 3 12 12",
   );
   $Centre_frame->configure(
       -borderwidth => 3,
   );
   $Centre_frame->g_grid(
       -column => 0,
       -row => 0,
       -sticky => "nsew",
   );



   $Main_window->g_grid_columnconfigure( 0, -weight => 1);
   $Main_window->g_grid_rowconfigure( 0, -weight => 1);
   $Container->g_grid_columnconfigure( 0, -weight => 1);
   $Container->g_grid_columnconfigure( 1, -weight => 1);
   $Container->g_grid_columnconfigure( 2, -weight => 1);
   $Container->g_grid_columnconfigure( 3, -weight => 1);
   $Container->g_grid_columnconfigure( 4, -weight => 1);
   $Container->g_grid_columnconfigure( 5, -weight => 1);
   $Container->g_grid_rowconfigure( 0, -weight => 0);
   $Container->g_grid_rowconfigure( 1, -weight => 4);
   $Container->g_grid_rowconfigure( 2, -weight => 0);


   $Centre_frame->g_grid_columnconfigure(0, -weight => 1);
   $Centre_frame->g_grid_rowconfigure(0, -weight => 1);








      $Action_Buttons_row = $#Run_mode_options + 2;

      $CANCEL_button = $Centre_frame->new_ttk__button(
          -width => "10",
          -text => "Cancel",
          -command => \&D_ui_cancelled
      );
      $CANCEL_button->g_grid(
          -column => 1,
          -row => $Action_Buttons_row
      );
#     $CANCEL_button->g_tooltip(
#         "End the program\nimmediately"
#     );


      $NEXT_button = $Centre_frame->new_ttk__button(
          -width => "10",
          -text => "Next >>",
          -takefocus => 1,
          -state => "active",
          -default => "active",
          -command => sub {




                push @Wizard_steps_visited, $Current_step_name;
                $Current_step_name = pop @Wizard_steps;
                _tidy_interface();
                _callback_step($Current_step_name);

          }
      );
      $NEXT_button->g_grid(
          -column => 2,
          -row => $Action_Buttons_row,
      );



      $BACK_button = $Centre_frame->new_ttk__button(
         -width => "10",
         -text => "<< Back",
         -state => "disabled",
         -default => "disabled"
      );
      $BACK_button->g_grid(
         -column => 0,
         -sticky => 'e',
         -row => $Action_Buttons_row
      );


      pad_children($Centre_frame);


   $Utility_frame = $Container->new_ttk__labelframe(
       -borderwidth => 3,
       -text => " Utility buttons ",
       -padding => "7",
   );
   $Utility_frame->g_grid(
       -column => 0,
       -row => 2,
       -pady => "10 0",
       -sticky => "nsew",
   );


   $LICENCE_VIEW_button = $Utility_frame->new_ttk__button(
       -text => "View Licence",
       -command => sub {
           Copy_licence_buffer();

           confirm("");
           confirm("Licence terms");
           confirm("");
           for my $licence_line (@Licence_buffer){
               chomp $licence_line;
               confirm($licence_line);
           }


       }
   );
   $LICENCE_VIEW_button->g_grid(
       -column => 0,
       -row => 2,
   );
#  $LICENCE_VIEW_button->g_tooltip(
#     "The licence will be displayed\nin the message window"
#  );


   $EPIC_VIEW_LOG_button = $Utility_frame->new_ttk__button(
       -text => "View FETA log",
       -state => "disabled",
       -default => "disabled",
       -command => sub {

           if ($^O =~ /WIN32/i ) {
           my $editor = $ENV{ EPIC_EDITOR } || 'notepad';
           system qq[ start /b "$editor" "$Log_file_specification" ];
           }
           else {
           my $editor = $ENV{ EPIC_EDITOR } || '/usr/bin/nedit';
           system "( $editor $Log_file_specification ) &";
           }


       }
   );
   $EPIC_VIEW_LOG_button->g_grid(
       -column => 1,
       -row => 2,
   );
#  $EPIC_VIEW_LOG_button->g_tooltip(
#      "When the Epic client runs, its logging output\nwill be saved automatically to a file."
#      . "\nClick to open the log file in an editor."
#  );


   $SAVE_button = $Utility_frame->new_ttk__button(
       -text => "Save FFQRUN log",
       -state => "active",
       -command => sub {

          my $ext = "ffr";
          my $file_type_filter = "{{$ext files} {.$ext}}";

          $Ffr_file_specification = Tkx::tk___getSaveFile(
              -title => 'Choose a log file e',
              -initialdir => $Default_log_directory,
              -filetypes => $file_type_filter,
              -parent => $Main_window,
          );
          if ($Ffr_file_specification) {
              confirm(
                  "FFQ_RUN Log file: $Ffr_file_specification"
              );
              open LOGFILE, ">$Ffr_file_specification";
              print LOGFILE @Messages_buffer;
              close LOGFILE;
              confirm("Writing to $Ffr_file_specification complete.");
          }
          else {
              confirm(
                  'No FFQ_RUN log file name was selected'
                  . ' and no log file was written'
              );
          }

       }
   );
   $SAVE_button->g_grid(
       -column => 2,
       -row => 2,
   );
#  $SAVE_button->g_tooltip(
#      "Save the information shown\nin the window above to a file"
#  );


   $SETUP_button = $Utility_frame->new_ttk__button(
       -text => "Setup",
       -command => sub {
           Display_setup();
       },
   );
   $SETUP_button->g_grid(
       -column => 5,
       -row => 2,
       -sticky => "e");
#  $SETUP_button->g_tooltip("Change and save settings");

   $Utility_frame->g_grid_columnconfigure(
       5,
       -weight => 3,
   );

   pad_children($Utility_frame);


   $Log_frame = $Container->new_ttk__labelframe(
       -text => " Messages from FETA ",
       -padding => "7",
   );
   $Log_frame->configure(
       -borderwidth => 3,
   );
   $Log_frame->g_grid(
       -column => 0,
       -row => 1,
       -pady => "10 0",
   );
   $Logging_window = $Log_frame->new_tk__listbox(
       -width => 80,
       -height => 10,
       -bg => "white",
   );
   my $scrollbar_y = $Log_frame->new_ttk__scrollbar(
       -command => [$Logging_window, "yview"],
       -orient => "vertical",
   );
   my $scrollbar_x = $Log_frame->new_ttk__scrollbar(
       -command => [$Logging_window, "xview"],
       -orient => "horizontal",
   );
   $Logging_window->configure(
      -xscrollcommand => [$scrollbar_x, "set"]
   );
   $Logging_window->configure(
      -yscrollcommand => [$scrollbar_y, "set"]
   );
   $Logging_window->g_grid(
       -column => 0,
       -row => 1,
   );
   $scrollbar_x->g_grid(
       -column => 0,
       -row => 2,
       -sticky => "we",
   );
   $scrollbar_y->g_grid(
       -column => 1,
       -row => 1,
       -sticky => "ns",
   );


   my $copyright_char = chr(0xa9);

   $DB::single=1;

   confirm( split("\n", <<"CREDITS") );
FETA is Copyright $copyright_char 2010 - 2012
University of Cambridge, Cambridge, UK.

Your use of these programs is subject to licence terms.
Click the \"View Licence\" button to review the licence.

Nutritional tables used by FETA are $copyright_char Crown Copyright
and are used under the terms of the Open Government Licence.

The website:
http://epic-norfolk.org.uk/index.php/for-researchers/feta-download/
contains the latest information about FETA

The authors request that the following acknowledgement
be made on publications that use the FETA software:

\"The EPIC-Norfolk study (DOI 10.22025/2019.10.105.00004) has received funding from the Medical Research Council (MR/N003284/1 and MC-UU_12015/1) and Cancer Research UK (C864/A14136).
We are grateful to all the participants who have been part of the project and to the many members of the study teams at the University of Cambridge who have enabled this research.\"

Please cite the following reference:

Mulligan et al. (2012), FETA: A software utility for processing food
frequency questionnaires (in progress)

Email: EPIC-Norfolk\@mrc-epid.cam.ac.uk with questions and comments
   \n
CREDITS

   Tkx::update();


   $Main_window->g_bind(
       "<Return>",
       sub {$NEXT_button->invoke();}
   );
   $Main_window->g_bind(
       "<Escape>",
       sub {$CANCEL_button->invoke();}
   );


   confirm("Preparing for a new run of FETA");



       if (!$ENV{EPIC_FETA_CLIENT}) {
           confirm("EPIC_FETA_CLIENT is not set");

           confirm(
               'Set the EPIC_FETA_CLIENT (FETA client location) environment variable for your platform'
           );
           confirm(
               'before you run the user interface. Please refer to the help file'
           );

           if ( $NEXT_button ) {
               $NEXT_button -> configure(
                   -state => 'disabled',
                   -default => 'disabled'
               );
           }
           else {
               confirm(
                   "(Attempt to disable NEXT button when it doesn't exist)"
               );
           }


           my $Epic_ffq_file_specification = Tkx::tk___getOpenFile(
               -title=>"Choose the FETA client executable to run",
               -parent => $Main_window,
           );
           if (!$Epic_ffq_file_specification) {
               confirm("Please press the cancel button when you have read these messages.");
               _callback_step( 'cancel' );
           }
           else {
               confirm(
                   "EPIC_FETA_CLIENT (FETA client) is set to "
                   . $Epic_ffq_file_specification
               );
               $EPIC_client_program = $Epic_ffq_file_specification;


               if ( $NEXT_button ) {
                   $NEXT_button -> configure(
                       -state => 'active',
                       -default => 'active',
                   );
               }
               else {
                   confirm(
                       "(Attempt to activate NEXT button when it doesn't exist)"
                   );
               }

           }



       }
       else {

       $DB::single=1;

           $EPIC_client_program = $ENV{EPIC_FETA_CLIENT};
           confirm("EPIC_FETA_CLIENT is set to " . $EPIC_client_program);


           open STATUS, qq[ "$EPIC_client_program" --version 2>&1 |]
               or confirm("Can't fork: $!");
           VERSION_LOOP: while (my $local_version = <STATUS>) {
               if ($local_version =~ /Epic_Client/){
                   chomp $local_version;
                   $Epic_ffq_version = $local_version;
                   confirm("FETA client reports version: " . $Epic_ffq_version);
                   last VERSION_LOOP;
               }
           }
           if ( !close STATUS ) {
               confirm("");
               confirm("There was a problem checking the FETA client version: $!");
               confirm("Perhaps the file selected was not executable");
               confirm("Check with a local guru to make sure FETA is seup correctly");
               confirm("You need to restart FETA");

               if ( $NEXT_button ) {
                   $NEXT_button -> configure(
                       -state => 'disabled',
                       -default => 'disabled'
                   );
               }
               else {
                   confirm(
                       "(Attempt to disable NEXT button when it doesn't exist)"
                   );
               }

               Tkx::tk___messageBox(
                   -message => 'FETA client executable not specificed correctly',
                   -detail => 'Set the EPIC_FETA_CLIENT (FETA client location) environment variable and run again',
                   -icon => 'error',
                   -title => 'FETA client error',
               );
               exit(0);
               #Display_setup();
           }
           else {
               confirm("FETA client program completed successfully");
           }



           if ( !$ENV{EPIC_DEFAULT_FOLDER} ) {
               confirm(
                   "Your default folder is not set, using home (~)."
               );
               confirm(
                   "Set EPIC_DEFAULT_FOLDER to point to the folder, or use setup"
               );
               #$Default_folder = "~";
               $Default_folder = $ENV{HOME};
           }
           else {
               confirm(
                   "EPIC_DEFAULT_FOLDER is set to " . $ENV{EPIC_DEFAULT_FOLDER}
               );
               $Default_folder = $ENV{EPIC_DEFAULT_FOLDER};
           }
           $Setup_file_specification = $Default_folder. '/feta.conf';
           if (!$ENV{EPIC_INPUT_FOLDER}) {
               $Input_folder = $Default_folder;
               $Default_input_directory = $Input_folder;
           }
           else {
              $Default_input_directory = $ENV{ EPIC_INPUT_FOLDER };
           }
           if (!$ENV{EPIC_OUTPUT_FOLDER}) {
               $Output_folder = $Default_folder;
               $Default_output_directory = $Output_folder;
           }
           else {
               $Default_output_directory = $ENV{ EPIC_OUTPUT_FOLDER };
           }
           if (!$ENV{EPIC_LOG_FOLDER}) {
               $Log_folder = $Default_folder;
               $Default_log_directory = $Log_folder;
           }
           else {
               $Default_log_directory = $ENV{ EPIC_LOG_FOLDER };
           }


           if (!$ENV{FETA_CLIENT_CONF}) {
               confirm("No FETA client configuration file has been specified");
           }
           else {
               $Client_config_file_specification = $ENV{ FETA_CLIENT_CONF };
           }



           read_setup();




           $Wizard_total_steps = @Wizard_steps;

           $Current_step_name = pop @Wizard_steps;
           $Wizard_step{$Current_step_name}->();

       }






   Tkx::MainLoop;

}




sub pad_children {
    foreach (Tkx::SplitList($_[0]->g_winfo_children)) {
        Tkx::grid_configure(
            $_,
            -padx => 5,
            -pady => 5,
        );
   }
}



sub D_ui_cancelled {
    exit(0);
}


sub confirm {
    my @messages = @_;

    $Logging_window->insert("end", @messages);
    $Logging_window->see("end");
    # Add it to the Messages_buffer
    push(@Messages_buffer, map{ $_ . "\n"} @messages);
}


sub flash {
    my $bg_colour = shift;
    my $fg_colour = shift;

    my $message = $_[0];
    $Logging_window->insert("end", $message);
    $Logging_window->see("end");
    $Logging_window->itemconfigure("end",
        -background => $bg_colour,
        -foreground => $fg_colour,
    );
    # Add it to the Messages_buffer
    push(@Messages_buffer, "$message\n");
}



sub _tidy_interface {
    for my $widget (@Widget_refs_to_go) {
        $widget->g_destroy;
    }
    @Widget_refs_to_go = ();
}


sub _callback_step {
    (my $function) = @_;

    if (!$Wizard_step{$function}) {
        confirm("Selector does not exist: $_[0]");
    }
    else {
        $Wizard_step{$function}->();
    }
}


sub ws_get_input_file {
    my $count_rows = 0;

    $Step_title = " "
        . ucfirst $Current_step_name
        . " of "
        . $Wizard_total_steps
        . " - "
        . $Wizard_step_title{$Current_step_name};

    $Centre_frame->configure(
        -text => $Step_title
    );

    my $lbl = $Centre_frame->new_ttk__label(
        -text => $Wizard_step_label{$Current_step_name},
    );
    $lbl->g_grid(
        -column => 0,
        -columnspan => 4,
        -row => $count_rows++,
        -sticky => "w",
    );
    push(@ Widget_refs_to_go, $lbl);


    if ( $Back_indicator{ $Current_step_name } ) {
        $BACK_button->configure(
            -state => 'active',
        );
    }
    else {
        $BACK_button->configure(
            -state => 'disabled',
    );
    }
    if ( $Forward_indicator{ $Current_step_name } ) {
        $NEXT_button->configure(
            -state => 'active',
        );
    }
    else {
        $NEXT_button->configure(
            -state => 'disabled',
        );
    }

    if (!$Input_file_specification){

        if ( $NEXT_button ) {
            $NEXT_button -> configure(
                -state => 'disabled',
                -default => 'disabled'
            );
        }
        else {
            confirm(
                "(Attempt to disable NEXT button when it doesn't exist)"
            );
        }

    }

    $BACK_button->configure(
        -state => "normal",
        -default => "normal",
        -command => sub {

                push @Wizard_steps, $Current_step_name;
                $Current_step_name = pop @Wizard_steps_visited;
                _tidy_interface();
                _callback_step($Current_step_name);

        }
    );

    $Input_widget = $Centre_frame->new_ttk__entry(
        -textvariable => \$Input_file_specification,
        -width => 50,
    );
    push @Widget_refs_to_go, $Input_widget;

    $BROWSE_button = $Centre_frame->new_ttk__button(
        -text => "Browse",
        -command => sub {

            my $file_type_filter = [
                        ['csv files',   '.csv'],
                        ['all files',   '*'],
                    ];

            $Input_file_specification = Tkx::tk___getOpenFile(
                -title=>"Choose an input file to process",
                -initialdir => $Default_input_directory,
                -filetypes => $file_type_filter,
                -parent => $Main_window,
            );



            if ($Input_file_specification) {
                confirm("Input file: " . $Input_file_specification);
                $NEXT_button->configure(
                    -state => "normal",
                    -default => "active",
                    -command => sub {
                        validate_input();

                            push @Wizard_steps_visited, $Current_step_name;
                            $Current_step_name = pop @Wizard_steps;
                            _tidy_interface();
                            _callback_step($Current_step_name);

                  }
               );

            }

        }
    );
    push @Widget_refs_to_go, $BROWSE_button;

    $Input_widget->g_grid(
        -column => 0,
        -columnspan => 2,
        -row => $count_rows,
        -sticky => "w",
    );
    $BROWSE_button->g_grid(
        -column => 3,
        -sticky => 'w',
        -row => $count_rows,
    );


        pad_children($Centre_frame);
#       $BACK_button->g_tooltip($Back_tip{$Current_step_name});
#       $NEXT_button->g_tooltip($Next_tip{$Current_step_name});

}



sub ws_set_output_file {
    my $count_rows = 0;

    $Step_title = " "
        . ucfirst $Current_step_name
        . " of "
        . $Wizard_total_steps
        . " - "
        . $Wizard_step_title{$Current_step_name};

    $Centre_frame->configure(
        -text => $Step_title
    );

    my $lbl = $Centre_frame->new_ttk__label(
        -text => $Wizard_step_label{$Current_step_name},
    );
    $lbl->g_grid(
        -column => 0,
        -columnspan => 4,
        -row => $count_rows++,
        -sticky => "w",
    );
    push(@ Widget_refs_to_go, $lbl);


    if ( $Back_indicator{ $Current_step_name } ) {
        $BACK_button->configure(
            -state => 'active',
        );
    }
    else {
        $BACK_button->configure(
            -state => 'disabled',
    );
    }
    if ( $Forward_indicator{ $Current_step_name } ) {
        $NEXT_button->configure(
            -state => 'active',
        );
    }
    else {
        $NEXT_button->configure(
            -state => 'disabled',
        );
    }


    if (!defined($Output_file_specification)) {

        if ( $NEXT_button ) {
            $NEXT_button -> configure(
                -state => 'disabled',
                -default => 'disabled'
            );
        }
        else {
            confirm(
                "(Attempt to disable NEXT button when it doesn't exist)"
            );
        }

    }


    $BACK_button->configure(
        -state => "normal",
        -default => "normal",
        -command => sub {

                push @Wizard_steps, $Current_step_name;
                $Current_step_name = pop @Wizard_steps_visited;
                _tidy_interface();
                _callback_step($Current_step_name);

        }
    );

    $Output_widget = $Centre_frame -> new_ttk__entry(
        -textvariable => \$Output_file_specification,
        -width => 50,
    );
    push @Widget_refs_to_go, $Output_widget;

    $BROWSE_button = $Centre_frame->new_ttk__button(
        -text => "Specify\nOutput File",
        -command => sub {

            # The output file type is ".ffq"
            my $ffq = 'ffq';
            my $csv = 'csv';
            my $file_type_filter = [
                        ['csv files',      '.csv'],
                        ['ffq files',      '.ffq'],
                        ['all files',       '*'],
                        ];
            $Output_file_specification = Tkx::tk___getSaveFile(
                -title => "Specify an output file name",
                -initialdir => $Default_output_directory,
                -filetypes => $file_type_filter,
                -parent => $Main_window,
            );


            if ($Output_file_specification) {
                confirm("Output file: " . $Output_file_specification);
                $NEXT_button->configure(
                    -state => "normal",
                    -default => "active",
                    -command => sub {

                            push @Wizard_steps_visited, $Current_step_name;
                            $Current_step_name = pop @Wizard_steps;
                            _tidy_interface();
                            _callback_step($Current_step_name);

                    }
                );
#               $NEXT_button->g_tooltip("Ready to run");
                $BACK_button->configure(
                    -state => "normal",
                    -default => "normal",
                    -command => sub {

                            push @Wizard_steps, $Current_step_name;
                            $Current_step_name = pop @Wizard_steps_visited;
                            _tidy_interface();
                            _callback_step($Current_step_name);

                    }
                );
#               $BACK_button->g_tooltip("Choose input file");
            }

        }
    );
    push @Widget_refs_to_go, $BROWSE_button;

    $Output_widget -> g_grid(
        -column => 0,
        -columnspan => 4,
        -row => $count_rows,
        -sticky => "w",
    );
    $BROWSE_button -> g_grid(
        -column => 3,
        -row => $count_rows,
    );
    pad_children($Centre_frame);


        pad_children($Centre_frame);
#       $BACK_button->g_tooltip($Back_tip{$Current_step_name});
#       $NEXT_button->g_tooltip($Next_tip{$Current_step_name});

}


sub ws_select_mode {

    $Step_title = " "
        . ucfirst $Current_step_name
        . " of "
        . $Wizard_total_steps
        . " - "
        . $Wizard_step_title{$Current_step_name};

    $Centre_frame->configure(
        -text => $Step_title,
    );

    my $count_rows = 0;

    my $lbl = $Centre_frame->new_ttk__label(
        -text => $Wizard_step_label{$Current_step_name},
    );
    $lbl->g_grid(
        -column => 0,
        -columnspan => 4,
        -row => $count_rows++,
        -sticky => "w",
    );
    push(@ Widget_refs_to_go, $lbl);

    for my $mode (@ Run_mode_options)
    {
        my $rb = $Centre_frame->new_ttk__radiobutton(
           -text => $Run_mode_labels{$mode},
           -variable => \$Mode_selection,
           -value => $mode,
        );
#       $rb->g_tooltip( $Run_mode_tips{$mode} );
        $rb->g_grid(
           -column => 0,
           -columnspan => 4,
           -row => $count_rows++,
           -sticky => "w",
        );
        push(@ Widget_refs_to_go, $rb);
    }



    if ( $Back_indicator{ $Current_step_name } ) {
        $BACK_button->configure(
            -state => 'active',
        );
    }
    else {
        $BACK_button->configure(
            -state => 'disabled',
    );
    }
    if ( $Forward_indicator{ $Current_step_name } ) {
        $NEXT_button->configure(
            -state => 'active',
        );
    }
    else {
        $NEXT_button->configure(
            -state => 'disabled',
        );
    }


        pad_children($Centre_frame);
#       $BACK_button->g_tooltip($Back_tip{$Current_step_name});
#       $NEXT_button->g_tooltip($Next_tip{$Current_step_name});

}


sub ws_introduction {
    my $count_rows = 0;

    $Step_title = " "
        . ucfirst $Current_step_name
        . " of "
        . $Wizard_total_steps
        . " - "
        . $Wizard_step_title{$Current_step_name};

    $Centre_frame->configure(
        -text => $Step_title,
    );

    my $lbl = $Centre_frame->new_ttk__label(
        -text => $Wizard_step_label{$Current_step_name},
    );
    $lbl->g_grid(
        -column => 0,
        -columnspan => 4,
        -row => $count_rows++,
        -sticky => "w",
    );
    push(@ Widget_refs_to_go, $lbl);


    if ( $Back_indicator{ $Current_step_name } ) {
        $BACK_button->configure(
            -state => 'active',
        );
    }
    else {
        $BACK_button->configure(
            -state => 'disabled',
    );
    }
    if ( $Forward_indicator{ $Current_step_name } ) {
        $NEXT_button->configure(
            -state => 'active',
        );
    }
    else {
        $NEXT_button->configure(
            -state => 'disabled',
        );
    }


        pad_children($Centre_frame);
#       $BACK_button->g_tooltip($Back_tip{$Current_step_name});
#       $NEXT_button->g_tooltip($Next_tip{$Current_step_name});

}


sub ws_run_epic_ffq {
    my $count_rows = 0;


    $Step_title = " "
        . ucfirst $Current_step_name
        . " of "
        . $Wizard_total_steps
        . " - "
        . $Wizard_step_title{$Current_step_name};

    $Centre_frame->configure(
        -text => $Step_title
    );

    my $lbl = $Centre_frame->new_ttk__label(
        -text => $Wizard_step_label{$Current_step_name},
    );
    $lbl->g_grid(
        -column => 0,
        -columnspan => 4,
        -row => $count_rows++,
        -sticky => "w",
    );
    push(@ Widget_refs_to_go, $lbl);


        my $progress_bar_ref = $Centre_frame -> new_ttk__progressbar(
            -orient => 'horizontal',
            -length => 200,
            -mode => 'indeterminate',
        );
        $progress_bar_ref -> g_grid(
            -column => 1,
            -columnspan => 4,
            -row => $count_rows,
            -sticky => 'we',
        );
        $progress_bar_ref -> start;


    if ( $Back_indicator{ $Current_step_name } ) {
        $BACK_button->configure(
            -state => 'active',
        );
    }
    else {
        $BACK_button->configure(
            -state => 'disabled',
    );
    }
    if ( $Forward_indicator{ $Current_step_name } ) {
        $NEXT_button->configure(
            -state => 'active',
        );
    }
    else {
        $NEXT_button->configure(
            -state => 'disabled',
        );
    }


        pad_children($Centre_frame);
#       $BACK_button->g_tooltip($Back_tip{$Current_step_name});
#       $NEXT_button->g_tooltip($Next_tip{$Current_step_name});


    my $feedback_frequency = $Setup_feedback_frequency;



    confirm("Ready to run...");
    open STATUS,  qq[ "$EPIC_client_program" ]
       . qq[ --config "$Client_config_file_specification" ]
       . qq[ --input "$Input_file_specification" ]
       . qq[ --output "$Output_file_specification" ]
       . qq[ --style $Mode_selection 2>&1 |]
        or confirm(
            "Can't fork: $!"
        );


    if (! ($Log_file_specification = make_epicffq_log_name()) ) {
        return 1;
    }

    confirm("Log file: $Log_file_specification");
    open $EPICFFQ_LOGFILE, '>', $Log_file_specification;
    if (@Licence_buffer ) {
        @Licence_buffer = ();
    }

    Copy_licence_buffer();
    print $EPICFFQ_LOGFILE @Licence_buffer;

#   DateTime -> DefaultLocale('en-GB');
#   my $date = DateTime -> now();
    my $tm = localtime;
    my $date = sprintf("%04d-%02d-%02dT%02d:%02d:%02d",$tm->year+1900, ($tm->mon)+1, $tm->mday,$tm->hour, $tm->min, $tm->sec);
    print $EPICFFQ_LOGFILE <<"RUN_PARAMS";


################################################################################
This file was generated by $Wizard_frame_title
on : $date

The parameters you selected for this run were:

mode of operation         : Calculate nutrients for individuals with output: $Run_mode_labels{$Mode_selection}
input file specification  : $Input_file_specification
output file specification : $Output_file_specification
log file specification    : $Log_file_specification

The FETA program version is: $Epic_ffq_version

Your parameter selection was used to run the FETA client like this:

$EPIC_client_program \
    --config $Client_config_file_specification
    --input $Input_file_specification
    --output $Output_file_specification
    --style $Mode_selection

Logging output from the original program was filtered and rewritten by
the FETA wizard.

################################################################################


RUN_PARAMS



    my $loop_counter = 0;
    my $total_counter = 0;


    RUN_LOOP: while ( <STATUS> ) {


        # log lines always get written to the logfile.
        Tkx::update();


        if (/invalid (frequency|portion)/i) {
            s/Error/Input warning/i;
        }
        elsif (/supplied no/i) {
            s/Error/Input warning/i;
        }
        elsif (/using default/i) {
            s/Note/Action taken/i;
        }
        elsif (/supplied invalid weight/i) {
            s/Error/Input warning/i;
            s/weight/response/i;
        }
        elsif (/unable to load nutrient/i) {
            s/Error/Database error/i;
            chomp( my $line = $_);
            flash( 'red', 'white', $line );
        }
        elsif (/questionaire/i) {
            s/Questionaire/Questionnaire/i;
            chomp( my $line = $_);
            $line =~ m{ processed\(([0-9]+)\)  }xms;
            my $diaries_processed = $1 // 0;
            if ( $diaries_processed > 0 ) {
                flash( 'green', 'black',
                    "$diaries_processed questionnaires were processed successfully.");
            }
            else {
                flash( 'red', 'white', 'No diaries were processed successfully.');
            }
        }
        elsif (/error/i) {
            chomp( my $line = $_);
            flash( 'red', 'white', $line );
        }

        print $EPICFFQ_LOGFILE $_;
        ++ $loop_counter;
        if ( $loop_counter > $feedback_frequency ) {
            Tkx::update();

            $loop_counter = 0;
        }
    }
    $progress_bar_ref -> stop;
    $progress_bar_ref -> g_destroy;
    if ( !close STATUS ) {
        confirm("");
        confirm("There was a problem running FETA client: $!");
        Tkx::tk___messageBox(
            -message => 'FETA client program gave a fatal error',
            -detail => 'Check the FETA client message in the messages panel',
            -icon => 'error',
            -title => 'FETA client error',
        );

        $NEXT_button -> configure(
            -state => "active",
            -text => "Quit",
            -default => 'active',
            -command => sub {
                exit(0);
            }
        );
        $CANCEL_button -> configure(
            -state => 'disabled',
            -default => 'disabled',
        );


    }
    else {
        confirm("FETA client program completed successfully");

            # Enable the view log button
            $EPIC_VIEW_LOG_button->configure(
                -state => "active",
                -default => "active"
            );
            $NEXT_button -> configure(
                -state => "active",
                -text => "Finish",
                -command => sub {
                    exit(0);
                }
            );
            $CANCEL_button -> configure(
                -state => 'disabled',
                -default => 'disabled',
            );



        $DB::single=1;

        if ( $Mode_selection eq 'spreadsheet' ) {
            confirm('Rewriting output file...');

            my @out_rows;
            my $out_csv = Text::CSV->new( { binary => 1 } )
                or die 'Cannot use CSV module to read output file: '
                    . Text::CSV->error_diag();

            open my $ out_fh, "<:encoding(utf8)", $Output_file_specification
                or die "$Output_file_specification: $!";
            my $out_headers = $out_csv->getline( $out_fh );
            while ( my $out_row = $out_csv->getline( $out_fh ) ) {
                push @out_rows, $out_row;
            }
            $out_csv->eof or $out_csv->error_diag();
            close $out_fh;


            foreach my $candidate_row ( @out_rows ) {
                if ( $participant_q_errors{$candidate_row->[0]} ){
                    push @$candidate_row,
                        $participant_q_errors{$candidate_row->[0]}->{error_count};
                }
                else {
                    push @$candidate_row, 0;
                }
            }


            my $conf_file;

            $DB::single=1;

            if (! open $conf_file, "<",  "$Client_config_file_specification" ) {
                confirm("Setup file ( $Client_config_file_specification ) not found");
                confirm('Unable to translate headers for output file');
                return 0;
            }
            chomp (my @conf_lines = <$conf_file>);
            close $conf_file;
            my $match_line_tag = '\A nutrients=';
            my @matched = grep m{$match_line_tag}xgms, @conf_lines;

            if ( $matched[0] =~ m{\A nutrients="([-a-zA-Z0-9_:/\\\.\s\(\)]+)" \z}xgms ) {
                $nutrients_file_specification = $1;
                confirm('Nutrients file for lookup: ' . $nutrients_file_specification );
            }
            else {
                confirm('Unable to parse nutrients file specification');
                confirm( $matched[0] );
            }


            my $nut_csv;
            my %nut_content;

            $nut_csv = Text::CSV->new( { binary => 1 } )
                or die 'Cannot use the CSV module: '
                    .Text::CSV->error_diag();
            open my $nut_fh, "<:encoding(utf8)", $nutrients_file_specification
                or die "$nutrients_file_specification: $!";
            my $nut_header = $nut_csv->getline( $nut_fh );
            while ( my $nut_row = $nut_csv->getline( $nut_fh ) ) {
                $nut_content{'NUTRIENT_'.$nut_row->[0]}
                     -> {'NAME'} = $nut_row->[1];
                $nut_content{'NUTRIENT_'.$nut_row->[0]}
                     -> {'UNITS'} = $nut_row->[2];
            }
            $nut_csv->eof or $nut_csv->error_diag();

            $DB::single=1;

            close $nut_fh;


            my @out_names;
            my @out_units;

            push @$out_headers, 'Missing ticks';
            foreach my $attribute ( @$out_headers ) {
                if ( $nut_content{$attribute} ) {
                    push @out_names, $nut_content{$attribute}
                        -> { 'NAME' };
                    push @out_units, $nut_content{$attribute}
                        -> { 'UNITS' };
                }
                else {
                    push @out_names, $attribute;
                    push @out_units, '';
                }
            }
            my $csv_status;
            my $out_names_str;
            my $out_units_str;

            $csv_status = $nut_csv -> combine( @out_names );
            if ( $csv_status ) {
                $out_names_str = $nut_csv -> string();
            }
            $csv_status = $nut_csv -> combine( @out_units );
            if ( $csv_status ) {
                $out_units_str = $nut_csv -> string();
            }



            open my $out_csv_fh, ">:encoding(utf8)", $Output_file_specification
                or die "$Output_file_specification: $!";
            print $out_csv_fh $out_names_str, "\n";
            print $out_csv_fh $out_units_str, "\n";

            my $out_rr;
            for my $rr ( @out_rows ) {
                $nut_csv -> combine( @$rr );
                $out_rr = $nut_csv -> string();
                print $out_csv_fh $out_rr, "\n";
            }
            close $out_csv_fh or die "$Output_file_specification: $!";

            confirm('Output file transformed and saved.');
        }

    }
    close $EPICFFQ_LOGFILE;

}


sub Copy_licence_buffer {
@Licence_buffer = <<"End_lines" =~ m/(.*\R)/g if !@Licence_buffer;
FETA is Copyright (c) 2010 - 2012
University of Cambridge, Cambridge, UK.

Nutritional composition tables used by FETA are (c) Crown Copyright
and are used under the terms of the Open Government Licence.

Your use of these programs is subject to the following licence terms.
By using these programs you accept the terms of the licence.

libEpic is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

libEpic is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

The website:
http://epic-norfolk.org.uk/index.php/for-researchers/feta-download/
contains the latest information about FETA

The authors request that the following acknowledgement
be made on publications that use the FETA software:

\"The EPIC-Norfolk study (DOI 10.22025/2019.10.105.00004) has received funding from the Medical Research Council (MR/N003284/1 and MC-UU_12015/1) and Cancer Research UK (C864/A14136).
We are grateful to all the participants who have been part of the project and to the many members of the study teams at the University of Cambridge who have enabled this research.\"

Please cite the following reference:

Mulligan et al. (2012), FETA: A software utility for processing food
frequency questionnaires (in progress)

Email: EPIC-Norfolk\@mrc-epid.cam.ac.uk with questions and comments
\n
End_lines
return;
}


sub Display_setup {

my $record_of_default_input_dir = $Default_input_directory;
my $record_of_default_output_dir = $Default_output_directory;
my $record_of_default_log_dir = $Default_log_directory;


my $setup_win_ref = $Main_window -> new_toplevel;
$setup_win_ref -> g_wm_title("ffqrun- Setup Window");
my $setup_container_ref =
    $setup_win_ref -> new_ttk__frame(
        -padding => '10',
    );
my $setup_frame_ref = $setup_container_ref -> new_ttk__labelframe(
    -text => 'Setup dialogue',
    -padding => '3 3 12 12',
);
$setup_frame_ref -> configure(
    -borderwidth => 3,
);
$setup_container_ref -> g_grid(
    -column => 0,
    -row => 0,
);
$setup_frame_ref -> g_grid(
    -column => 0,
    -row => 0,
    -sticky => "nsew",
);



my $default_input_directory_label_ref =
    $setup_frame_ref -> new_ttk__label(
        -text => 'Input files are in folder:'
    );
my $default_input_directory_widget_ref =
    $setup_frame_ref -> new_ttk__entry(
        -textvariable => \$Default_input_directory,
        -width => 50,
    );

my $browse_in_dir_button_ref =
    $setup_frame_ref -> new_ttk__button(
        -text => 'Browse',
        -command => sub {

            $Default_input_directory =
                Tkx::tk___chooseDirectory(
                    -title => "Folder for input files",
                    -parent => $setup_win_ref,
                    -initialdir => $Default_input_directory,
            );

        },
    );
$default_input_directory_label_ref -> g_grid(
    -column => 1,
    -columnspan => 5,
    -row => 1,
    -sticky => 'w',
);
$default_input_directory_widget_ref -> g_grid(
    -column => 1,
    -columnspan => 4,
    -row => 2,
    -sticky => "w",
);
$browse_in_dir_button_ref -> g_grid(
    -column => 5,
    -row => 2,
);


my $default_output_directory_label_ref =
    $setup_frame_ref -> new_ttk__label(
        -text => 'Output files are in folder:'
    );
my $default_output_directory_widget_ref =
    $setup_frame_ref -> new_ttk__entry(
        -textvariable => \$Default_output_directory,
        -width => 50,
    );

my $browse_out_dir_button_ref =
    $setup_frame_ref -> new_ttk__button(
        -text => 'Browse',
        -command => sub {

            $Default_output_directory =
                Tkx::tk___chooseDirectory(
                    -title => "Folder for output files",
                    -parent => $setup_win_ref,
                    -initialdir => $Default_output_directory,
            );

        },
    );
$default_output_directory_label_ref -> g_grid(
    -column => 1,
    -columnspan => 5,
    -row => 3,
    -sticky => 'w',
);
$default_output_directory_widget_ref -> g_grid(
    -column => 1,
    -columnspan => 4,
    -row => 4,
    -sticky => "w",
);
$browse_out_dir_button_ref -> g_grid(
    -column => 5,
    -row => 4,
);


my $default_log_directory_label_ref =
    $setup_frame_ref -> new_ttk__label(
        -text => 'Log files are in folder:'
    );
my $default_log_directory_widget_ref =
    $setup_frame_ref -> new_ttk__entry(
        -textvariable => \$Default_log_directory,
        -width => 50,
    );

my $browse_log_dir_button_ref =
    $setup_frame_ref -> new_ttk__button(
        -text => 'Browse',
        -command => sub {

            $Default_log_directory =
                Tkx::tk___chooseDirectory(
                    -title => "Folder for log files",
                    -parent => $setup_win_ref,
                    -initialdir => $Default_log_directory,
            );

        },
    );
$default_log_directory_label_ref -> g_grid(
    -column => 1,
    -columnspan => 5,
    -row => 5,
    -sticky => 'w',
);
$default_log_directory_widget_ref -> g_grid(
    -column => 1,
    -columnspan => 4,
    -row => 6,
    -sticky => 'w',
);
$browse_log_dir_button_ref -> g_grid(
    -column => 5,
    -row => 6,
);




my $ok_setup_button_ref =
    $setup_frame_ref -> new_ttk__button(
        -width => '10',
        -text => 'OK',
        -state => 'active',
        -default => 'active',
        -command => sub {
            # New settings will be saved by default
            $setup_win_ref -> g_destroy;
        },
    );
$ok_setup_button_ref -> g_grid(
    -column => 4,
    -row => 10,
);
my $cancel_setup_button_ref =
    $setup_frame_ref -> new_ttk__button(
        -width => '10',
        -text => 'Cancel',
        -state => 'active',
        -default => 'normal',
        -command => sub {


            $Default_input_directory = $record_of_default_input_dir || '~';
            $Default_output_directory = $record_of_default_output_dir || '~';
            $Default_log_directory = $record_of_default_log_dir || '~';

            $setup_win_ref -> g_destroy;
        },
    );
$cancel_setup_button_ref -> g_grid(
    -column =>3,
    -row => 10,
);

my $save_setup_button_ref =
    $setup_frame_ref -> new_ttk__button(
        -width => '10',
        -text => 'Save',
        -state => 'active',
        -default => 'active',
        -command => sub {
            # New settings will be saved
            save_setup();
            confirm("Settings saved.");
        },
    );
$save_setup_button_ref -> g_grid(
    -column => 2,
    -row => 10,
);




pad_children($setup_frame_ref);

return;
}


sub log_fail {
        confirm('Unable to access the default log directory');
        confirm('Default log directory is: ' . $Default_log_directory);
        confirm('Use setup to change the default log directory');

        $NEXT_button -> configure(
            -state => "active",
            -text => "Quit",
            -default => 'active',
            -command => sub {
                exit(0);
            }
        );
        $CANCEL_button -> configure(
            -state => 'disabled',
            -default => 'disabled',
        );


        return(1);
} ;


sub make_epicffq_log_name {
    my $local_error;

    opendir (my $dir_handle, $Default_log_directory) ||
        ($local_error = log_fail());

    if ($local_error){
        return(0);
    }

    my @log_names = grep { -f "$Default_log_directory/$_" } readdir($dir_handle);
    closedir $dir_handle;

    my $stripped_output_name;

    if ( $Output_file_specification =~ / \A (.*?) [.].*? \z /xms ) {
        $stripped_output_name = $1 // 'ffq';
    }
    if ( $stripped_output_name =~ qr{ \A .*/ (.*?) \z }xms ) {
        $stripped_output_name = $1 // 'ffqrun';
    }
    my $partial_logfile_name = $Default_log_directory
        . '/'
        . $stripped_output_name;


#   DateTime -> DefaultLocale('en-GB');
#   my $date = DateTime -> now();
#   # No separator
#   my $ymd = $date -> ymd('');
my $tm = localtime;
my $date = sprintf("%04d-%02d-%02dT%02d:%02d:%02d",$tm->year+1900, ($tm->mon)+1, $tm->mday,$tm->hour, $tm->min, $tm->sec);
my $ymd = sprintf("%04d%02d%02d",$tm->year+1900, ($tm->mon)+1, $tm->mday);

    my $sequence_num =      qr{ [[:digit:]]*    }xms;
    my $extension =         qr{ [.]log          }xms;
    my $seq_no_matched =    qr{ $stripped_output_name
                                -
                                $ymd
                                -
                                ($sequence_num)
                                $extension      }xms;
# Count the number of matches



my @seqnums = ();
foreach my $name ( @log_names ) {
    if ( $name =~ $seq_no_matched ) {
        push @seqnums, $1;
    }
}

my $log_sequence;
$log_sequence = max @seqnums;
my $log_output_name =
    $partial_logfile_name
        . '-'
        . $ymd
        . '-'
        . ++ $log_sequence
        . '.log';
return $log_output_name;
}


sub ws_cancel_ffq {
    _tidy_interface();

    my $Current_step_name = 'cancel';
    $Step_title = ' '
        . $Wizard_step_title{ 'cancel' };
    $Centre_frame->configure(
        -text => $Step_title
    );

    if ( $NEXT_button ) {
        $NEXT_button -> configure(
            -state => 'disabled',
            -default => 'disabled'
        );
    }
    else {
        confirm(
            "(Attempt to disable NEXT button when it doesn't exist)"
        );
    }


    if ( $Back_indicator{ $Current_step_name } ) {
        $BACK_button->configure(
            -state => 'active',
        );
    }
    else {
        $BACK_button->configure(
            -state => 'disabled',
    );
    }
    if ( $Forward_indicator{ $Current_step_name } ) {
        $NEXT_button->configure(
            -state => 'active',
        );
    }
    else {
        $NEXT_button->configure(
            -state => 'disabled',
        );
    }



        pad_children($Centre_frame);
#       $BACK_button->g_tooltip($Back_tip{$Current_step_name});
#       $NEXT_button->g_tooltip($Next_tip{$Current_step_name});

}


sub validate_input {

    use Text::CSV;

    my @rows;
    my $csv = Text::CSV->new( { binary => 1 } )
        or die 'Cannot use CSV module: '
            .Text::CSV->error_diag();

    open my $fh, "<:encoding(utf8)", $Input_file_specification
        or die "$Input_file_specification: $!";
    my $headers = $csv->getline( $fh );
    while( my $row = $csv->getline( $fh ) ) {
        push @rows, $row;
    }
    $csv->eof or $csv->error_diag();
    close $fh;


    my %seen = ();

    OUTER_ROW:foreach my $row_ref( @rows ) {
        next OUTER_ROW if $seen{ $row_ref->[0] }++;
    }
    foreach my $id (keys %seen) {
        if ($seen{$id} > 1) {
            flash( 'red', 'white',
             "Input Validation Error: ID $id was seen $seen{$id} times"
            );
        }
    }



    my %ignore_fields = (
        ID => 1,
        MILK_FOOD => 1,
        CEREAL_FOOD => 1,
        FAT_FRYING_FOOD => 1,
        FAT_BAKING_FOOD => 1,
        ZZA => 1,
        ZZB => 1,
        ZZC => 1,
        ZZD => 1,
        GRILL_FREQUENCY => 1,
        HOW_COOKED => 1,
        ADD_SALT_COOKING => 1,
        ADD_SALT_TABLE => 1,
        SALT_SUBSTITUTE => 1,
        DE_INITIALS => 1,
        VERSION => 1,
        DATE_OF_ENTRY => 1,
        MISSING_TICKS => 1,
        SEX => 1,
        DOC => 1,
        VEG => 1,
        SALAD => 1,
        FRUIT => 1,
        FISH => 1,
        MEAT => 1,
        DOB =>1,
    );


    my %allowed_fields = (
    BEEF => 1, BURGER => 1, PORK => 1, LAMB => 1, CHICKEN => 1,
    BACON => 1, HAM => 1, CORNED_BEEF => 1, SAUSAGES => 1,
    SAVOURY_PIES => 1, LIVER => 1, FRIED_FISH => 1, FISH_FINGERS
    => 1, WHITE_FISH => 1, OILY_FISH => 1, SHELLFISH => 1, ROE
    => 1, WHITE_BREAD => 1, BROWN_BREAD => 1, WHOLEMEAL_BREAD =>
    1, CRACKERS => 1, CRISPBREAD => 1, PORRIDGE => 1, CEREAL =>
    1, BOILED_POTATOES => 1, CHIPS => 1, ROAST_POTATOES => 1,
    POTATO_SALAD => 1, WHITE_RICE => 1, BROWN_RICE => 1,
    WHITE_PASTA => 1, WHOLEMEAL_PASTA => 1, LASAGNE=> 1, PIZZA=>
    1, SINGLE_CREAM => 1, DOUBLE_CREAM => 1, LOWFAT_YOGURT => 1,
    FULLFAT_YOGURT=> 1, DAIRY_DESSERT => 1, CHEESE => 1,
    COTTAGE_CHEESE => 1, EGGS => 1, QUICHE => 1,
    LOWCAL_SALAD_CREAM => 1, SALAD_CREAM => 1, FRENCH => 1,
    OTHER_DRESSING => 1, BUTTER => 1, HARD_MARGARINE => 1,
    POLYUNSATURATED_MARGARINE => 1, OTHER_MARGARINE => 1,
    LOWFAT_SPREAD => 1, VERY_LOWFAT_SPREAD => 1,
    CHOCOLATE_BISCUIT => 1, PLAIN_BISCUIT => 1, HOMEBAKED_CAKE
    => 1, READYMADE_CAKE => 1, HOMEBAKED_BUNS => 1,
    READYMADE_BUNS => 1, HOMEBAKED_FRUIT_PIES => 1,
    READYMADE_FRUIT_PIES => 1, HOMEBAKED_SPONGE => 1,
    READYMADE_SPONGE => 1, MILK_PUDDINGS => 1, ICE_CREAM => 1,
    CHOCOLATES => 1, CHOCOLATE_BARS => 1, SWEETS => 1, SUGAR =>
    1, CRISPS => 1, NUTS => 1, VEGETABLE_SOUP => 1, MEAT_SOUP =>
    1, SAUCES => 1, KETCHUP => 1, PICKLES => 1, MARMITE => 1,
    JAM => 1, PEANUT_BUTTER => 1, TEA => 1, INSTANT_COFFEE => 1,
    DECAFF_COFFEE => 1, COFFEE_WHITENER => 1, COCOA => 1,
    HORLICKS => 1, WINE => 1, BEER => 1, PORT => 1, SPIRITS =>
    1, LOWCAL_FIZZY_DRINKS => 1, FIZZY_DRINKS => 1, FRUIT_JUICE
    => 1, FRUIT_SQUASH => 1, APPLES => 1, PEARS => 1, ORANGES =>
    1, GRAPEFRUIT => 1, BANANAS => 1, GRAPES => 1, MELONS => 1,
    PEACHES => 1, STRAWBERRIES => 1, TINNED_FRUIT => 1,
    DRIED_FRUIT => 1, CARROTS => 1, SPINACH => 1, BROCCOLI => 1,
    SPROUTS => 1, CABBAGE => 1, PEAS => 1, GREEN_BEANS => 1,
    MARROW => 1, CAULIFLOWER => 1, PARSNIPS => 1, LEEKS => 1,
    ONIONS => 1, GARLIC => 1, MUSHROOMS => 1, PEPPERS => 1,
    BEANSPROUTS => 1, GREEN_SALAD => 1, WATERCRESS => 1,
    TOMATOES => 1, SWEETCORN => 1, BEETROOT => 1, COLESLAW => 1,
    AVOCADO => 1, BEANS => 1, LENTILS => 1, TOFU => 1,);

    my $line = 1; # 0th line is the header line
    OUTER: foreach my $data_row ( @rows ) {
        ++$line;
        my $i = 0;
        INNER: foreach my $datum ( @$data_row ) {
            next INNER if defined ( $ignore_fields{$headers->[$i++]} );
            if ( $datum =~ m{\A -4|-5 \z}xgms ) {

                my $numeric_column = $i;
                my $text_column = numeric_to_text( $numeric_column );

                flash( 'yellow', 'blue',
                    'Input line no. '
                    . $line
                    . ' Validation Warning: '
                    . $data_row->[0]
                    . ' field('
                    . $numeric_column
                    . ' - '
                    . $text_column
                    . '): '
                    . $headers->[$numeric_column - 1]
                    . ' check frequency: '
                    . $datum
                );


                if ( $allowed_fields{ $headers->[$i] } ) {
                    $participant_q_errors{ $data_row->[0] }
                        -> {error_count}++;
                }

                next INNER;
            }
            if ( $datum =~ m{\A -1|-9 \z}xgms ) {

                if ( $allowed_fields{ $headers->[$i] } ) {
                    $participant_q_errors{ $data_row->[0] }
                        -> {error_count}++;
                }

                next INNER;
            }
            if ( $datum =~ m{\A [1-9] ; [1-9] \z}xgms ) {

                $mult_freq{ $data_row->[0] }
                    -> {line} = $line - 1;
                $mult_freq{ $data_row->[0] }
                    -> {column} = $i - 1;
                $mult_freq{ $data_row->[0] }
                    -> {frequencies} = $datum;


                my $numeric_column = $i;
                my $text_column = numeric_to_text( $numeric_column );

                flash( 'green', 'black',
                    'Input line no. '
                    . $line
                    . ' Validation Warning: '
                    . $data_row->[0]
                    . ' field('
                    . $numeric_column
                    . ' - '
                    . $text_column
                    . '): '
                    . $headers->[$numeric_column - 1]
                    . ' multiple frequencies: '
                    . $datum
                );



                next INNER;
            }
            if (  $datum !~ m{
                    (\A -1 \z) | (\A -9 \z) | (\A [1-9] \z)
                }xgms )  {
                my $col = $i +1;

                my $numeric_column = $i;
                my $text_column = numeric_to_text( $numeric_column );


                if ($Count_reds{ $headers->[$numeric_column - 1] } ++ >= $Red_limit) {
                    $ignore_fields{ $headers->[$numeric_column - 1] } = 1;
                    flash( 'purple', 'white',
                        'Input Validation Warning: field('
                        . $numeric_column
                        . ' - '
                        . $text_column
                        . '): '
                        . $headers->[$numeric_column - 1]
                        . ' has more than '
                        . $Red_limit
                        . ' frequency errors and will now be ignored'
                    );

                    next INNER;
                }


                flash( 'red', 'white',
                    'Input line no. '
                    . $line
                    . ' Validation Warning: '
                    . $data_row->[0]
                    . ' field('
                    . $numeric_column
                    . ' - '
                    . $text_column
                    . '): '
                    . $headers->[$numeric_column - 1]
                    . ' invalid frequency: '
                    . $datum
                );


                if ( $allowed_fields{ $headers->[$i] } ) {
                    $participant_q_errors{ $data_row->[0] }
                        -> {error_count}++;
                }

                next INNER;
            }
        }
        if ( $participant_q_errors{ $data_row->[0] } ) {
            if ( $participant_q_errors{ $data_row->[0] }->{error_count} > 10 ) {

                my $numeric_column = $i;
                my $text_column = numeric_to_text( $numeric_column );

                flash( 'cornflower blue', 'white',
                    'Participant ID: '
                    . $data_row->[0]
                    . ': Too many missing ticks: '
                    . $participant_q_errors{ $data_row->[0] }->{error_count}
                );

            }
        }
    }




    my %master;

    my @fat_codes = qw(
    17013 17014 17031 a7026 17004 17005 17033 17034 a7022 17017 17015 17024 17025
    17023 17028 17027 17026 17007 17008 17009 17036 17037 17010 12259 17020 17019
    17018 17038 17039 17040 17021 17041 17042 17043 a7021 17044 17045 17046 17047
    );


    my @milk_codes = qw(
    12024 12018 a2003 a2001 12037 a2002 -1 12008 12041 12012 12001 12001 12042
    12130
    );


    my @cereal_codes = qw(
    11126 11127 11128 a1362 11129 68 11130 a1361 11131 11133 11134 11135 11136
    11137 11140 75 11144 79 11146 11147 11148 11149 11150 11151 11152 11153 11154
    11155 11157 11034
    );


    foreach my $value ( @milk_codes ) {
        $master{ $value } = 'milk';
    }
    foreach my $value ( @cereal_codes ) {
        $master{ $value } = 'cereal';
    }
    foreach my $value ( @fat_codes ) {
        $master{ $value } = 'fat';
    }


    my %index;

    $index{milk_food} = iarray('MILK_FOOD', @$headers);
    $index{cereal_food} = iarray('CEREAL_FOOD', @$headers);
    $index{fat_frying_food} = iarray('FAT_FRYING_FOOD', @$headers);
    $index{fat_baking_food} = iarray('FAT_BAKING_FOOD', @$headers);

    $line = 1;
    my $i;
    my @foods = keys (%index);
    foreach my $data_row ( @rows ) {
        ++$line;
        FOOD_CODE: foreach my $food ( @foods ) {
            $i = $index{ $food };
            my $data = $data_row->[ $i ];
            my @codes = split( ';', $data );
            foreach my $code ( @codes ) {
                if (! defined( $master{$code} )) {

                    my $numeric_column = $i;
                    my $text_column = numeric_to_text( $numeric_column );

                    flash( 'red', 'white',
                        'Input line no. '
                        . $line
                        . ' Validation Warning: '
                        . $data_row->[0]
                        . ' field('
                        . $numeric_column
                        . ' - '
                        . $text_column
                        . '): '
                        . $headers->[$numeric_column]
                        . ' unrecognized food code: '
                        . $code
                    );

                    next FOOD_CODE;
                }
                else {
                    # check code is in the correct column



                    next FOOD_CODE;
                }
            }
        }
    }

    return(1);
}


sub numeric_to_text{
    my $dividend = shift;

    my $column_name = '';
    my $modulo;

    while ($dividend > 0)
    {
        $modulo = ($dividend - 1) % 26;
        $column_name = chr(65 + $modulo) . $column_name;
        $dividend = int(($dividend - $modulo) / 26);
    }

    return $column_name;
}


sub iarray {
    my $item = shift;

    my $i = 0;
    foreach my $candidate ( @_ ) {
        if ( $candidate eq $item ) {
            return $i;
        }
        ++$i;
    }
    return -1;
}


sub read_setup {
my $setup_file_handle;


$DB::single=1;


if (! open $setup_file_handle, "<", $Setup_file_specification) {
    confirm("Setup file ( $Setup_file_specification ) not found");
    return 0;
}

load_setup_file( $setup_file_handle );
close $setup_file_handle;
confirm("Settings were read from $Setup_file_specification");
return 1;
}


sub shell_quote {
    my $item_string = shift;

    return "'" . $item_string . "'";

}


sub save_setup{
my $setup_file_handle;


$DB::single=1;


if (! open $setup_file_handle, ">", $Setup_file_specification) {
    confirm(
"Setup file ( $Setup_file_specification ) could not be opened for write access"
    );
    return 0;
}
dump_setup_file( $setup_file_handle, $actions );
close $setup_file_handle;
return 1;
}


sub dump_setup_file {
    my $fh = shift;
    my $action_table = shift;

    my @setup_keys = keys %$action_table;
    foreach my $key_out ( @setup_keys ) {
        next if $key_out =~ /_DEFAULT_/ ;
        print $fh $key_out.' '.$action_table -> {$key_out}-> ($key_out)."\n";
    }
}


sub load_setup_file {
    my $file_handle = shift;

    while ( <$file_handle> ) {
        chomp;
        my ($directive, $rest) = split /\s+/, $_, 2;
        my $action = $actions->{$directive} || $actions->{_DEFAULT_};
        if ($action) {
            $action->($directive, $rest);
        }
        else {
            confirm(
        "Unrecognised directive $directive on line $. of config_file"
            );
        }
    }


}
